# Copyright 2014-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

"""Tools to parse mongo client options."""

from bson.codec_options import _parse_codec_options
from pymongo.auth import _build_credentials_tuple
from pymongo.common import validate_boolean
from pymongo import common
from pymongo.compression_support import CompressionSettings
from pymongo.errors import ConfigurationError
from pymongo.monitoring import _EventListeners
from pymongo.pool import PoolOptions
from pymongo.read_concern import ReadConcern
from pymongo.read_preferences import (make_read_preference,
                                      read_pref_mode_from_name)
from pymongo.server_selectors import any_server_selector
from pymongo.ssl_support import get_ssl_context
from pymongo.write_concern import WriteConcern


def _parse_credentials(username, password, database, options):
    """Parse authentication credentials."""
    mechanism = options.get('authmechanism', 'DEFAULT' if username else None)
    source = options.get('authsource')
    if username or mechanism:
        return _build_credentials_tuple(
            mechanism, source, username, password, options, database)
    return None


def _parse_read_preference(options):
    """Parse read preference options."""
    if 'read_preference' in options:
        return options['read_preference']

    name = options.get('readpreference', 'primary')
    mode = read_pref_mode_from_name(name)
    tags = options.get('readpreferencetags')
    max_staleness = options.get('maxstalenessseconds', -1)
    return make_read_preference(mode, tags, max_staleness)


def _parse_write_concern(options):
    """Parse write concern options."""
    concern = options.get('w')
    wtimeout = options.get('wtimeoutms')
    j = options.get('journal')
    fsync = options.get('fsync')
    return WriteConcern(concern, wtimeout, j, fsync)


def _parse_read_concern(options):
    """Parse read concern options."""
    concern = options.get('readconcernlevel')
    return ReadConcern(concern)


def _parse_ssl_options(options):
    """Parse ssl options."""
    use_tls = options.get('tls')
    if use_tls is not None:
        validate_boolean('tls', use_tls)

    certfile = options.get('tlscertificatekeyfile')
    passphrase = options.get('tlscertificatekeyfilepassword')
    ca_certs = options.get('tlscafile')
    crlfile = options.get('tlscrlfile')
    allow_invalid_certificates = options.get('tlsallowinvalidcertificates', False)
    allow_invalid_hostnames = options.get('tlsallowinvalidhostnames', False)
    disable_ocsp_endpoint_check = options.get('tlsdisableocspendpointcheck', False)

    enabled_tls_opts = []
    for opt in ('tlscertificatekeyfile', 'tlscertificatekeyfilepassword',
                'tlscafile', 'tlscrlfile'):
        # Any non-null value of these options implies tls=True.
        if opt in options and options[opt]:
            enabled_tls_opts.append(opt)
    for opt in ('tlsallowinvalidcertificates', 'tlsallowinvalidhostnames',
                'tlsdisableocspendpointcheck'):
        # A value of False for these options implies tls=True.
        if opt in options and not options[opt]:
            enabled_tls_opts.append(opt)

    if enabled_tls_opts:
        if use_tls is None:
            # Implicitly enable TLS when one of the tls* options is set.
            use_tls = True
        elif not use_tls:
            # Error since tls is explicitly disabled but a tls option is set.
            raise ConfigurationError("TLS has not been enabled but the "
                                     "following tls parameters have been set: "
                                     "%s. Please set `tls=True` or remove."
                                     % ', '.join(enabled_tls_opts))

    if use_tls:
        ctx = get_ssl_context(
            certfile,
            passphrase,
            ca_certs,
            crlfile,
            allow_invalid_certificates,
            allow_invalid_hostnames,
            disable_ocsp_endpoint_check)
        return ctx, allow_invalid_hostnames
    return None, allow_invalid_hostnames


def _parse_pool_options(options):
    """Parse connection pool options."""
    max_pool_size = options.get('maxpoolsize', common.MAX_POOL_SIZE)
    min_pool_size = options.get('minpoolsize', common.MIN_POOL_SIZE)
    max_idle_time_seconds = options.get(
        'maxidletimems', common.MAX_IDLE_TIME_SEC)
    if max_pool_size is not None and min_pool_size > max_pool_size:
        raise ValueError("minPoolSize must be smaller or equal to maxPoolSize")
    connect_timeout = options.get('connecttimeoutms', common.CONNECT_TIMEOUT)
    socket_timeout = options.get('sockettimeoutms')
    wait_queue_timeout = options.get(
        'waitqueuetimeoutms', common.WAIT_QUEUE_TIMEOUT)
    event_listeners = options.get('event_listeners')
    appname = options.get('appname')
    driver = options.get('driver')
    server_api = options.get('server_api')
    compression_settings = CompressionSettings(
        options.get('compressors', []),
        options.get('zlibcompressionlevel', -1))
    ssl_context, tls_allow_invalid_hostnames = _parse_ssl_options(options)
    load_balanced = options.get('loadbalanced')
    max_connecting = options.get('maxconnecting', common.MAX_CONNECTING)
    return PoolOptions(max_pool_size,
                       min_pool_size,
                       max_idle_time_seconds,
                       connect_timeout, socket_timeout,
                       wait_queue_timeout,
                       ssl_context, tls_allow_invalid_hostnames,
                       _EventListeners(event_listeners),
                       appname,
                       driver,
                       compression_settings,
                       max_connecting=max_connecting,
                       server_api=server_api,
                       load_balanced=load_balanced)


class ClientOptions(object):
    """Read only configuration options for a MongoClient.

    Should not be instantiated directly by application developers. Access
    a client's options via :attr:`pymongo.mongo_client.MongoClient.options`
    instead.
    """

    def __init__(self, username, password, database, options):
        self.__options = options

        self.__codec_options = _parse_codec_options(options)
        self.__credentials = _parse_credentials(
            username, password, database, options)
        self.__direct_connection = options.get('directconnection')
        self.__local_threshold_ms = options.get(
            'localthresholdms', common.LOCAL_THRESHOLD_MS)
        # self.__server_selection_timeout is in seconds. Must use full name for
        # common.SERVER_SELECTION_TIMEOUT because it is set directly by tests.
        self.__server_selection_timeout = options.get(
            'serverselectiontimeoutms', common.SERVER_SELECTION_TIMEOUT)
        self.__pool_options = _parse_pool_options(options)
        self.__read_preference = _parse_read_preference(options)
        self.__replica_set_name = options.get('replicaset')
        self.__write_concern = _parse_write_concern(options)
        self.__read_concern = _parse_read_concern(options)
        self.__connect = options.get('connect')
        self.__heartbeat_frequency = options.get(
            'heartbeatfrequencyms', common.HEARTBEAT_FREQUENCY)
        self.__retry_writes = options.get('retrywrites', common.RETRY_WRITES)
        self.__retry_reads = options.get('retryreads', common.RETRY_READS)
        self.__server_selector = options.get(
            'server_selector', any_server_selector)
        self.__auto_encryption_opts = options.get('auto_encryption_opts')
        self.__load_balanced = options.get('loadbalanced')

    @property
    def _options(self):
        """The original options used to create this ClientOptions."""
        return self.__options

    @property
    def connect(self):
        """Whether to begin discovering a MongoDB topology automatically."""
        return self.__connect

    @property
    def codec_options(self):
        """A :class:`~bson.codec_options.CodecOptions` instance."""
        return self.__codec_options

    @property
    def _credentials(self):
        """A :class:`~pymongo.auth.MongoCredentials` instance or None."""
        return self.__credentials

    @property
    def direct_connection(self):
        """Whether to connect to the deployment in 'Single' topology."""
        return self.__direct_connection

    @property
    def local_threshold_ms(self):
        """The local threshold for this instance."""
        return self.__local_threshold_ms

    @property
    def server_selection_timeout(self):
        """The server selection timeout for this instance in seconds."""
        return self.__server_selection_timeout

    @property
    def server_selector(self):
        return self.__server_selector

    @property
    def heartbeat_frequency(self):
        """The monitoring frequency in seconds."""
        return self.__heartbeat_frequency

    @property
    def pool_options(self):
        """A :class:`~pymongo.pool.PoolOptions` instance."""
        return self.__pool_options

    @property
    def read_preference(self):
        """A read preference instance."""
        return self.__read_preference

    @property
    def replica_set_name(self):
        """Replica set name or None."""
        return self.__replica_set_name

    @property
    def write_concern(self):
        """A :class:`~pymongo.write_concern.WriteConcern` instance."""
        return self.__write_concern

    @property
    def read_concern(self):
        """A :class:`~pymongo.read_concern.ReadConcern` instance."""
        return self.__read_concern

    @property
    def retry_writes(self):
        """If this instance should retry supported write operations."""
        return self.__retry_writes

    @property
    def retry_reads(self):
        """If this instance should retry supported read operations."""
        return self.__retry_reads

    @property
    def auto_encryption_opts(self):
        """A :class:`~pymongo.encryption.AutoEncryptionOpts` or None."""
        return self.__auto_encryption_opts

    @property
    def load_balanced(self):
        """True if the client was configured to connect to a load balancer."""
        return self.__load_balanced

    @property
    def event_listeners(self):
        """The event listeners registered for this client.

        See :mod:`~pymongo.monitoring` for details.

        .. versionadded:: 4.0
        """
        return self.__pool_options._event_listeners.event_listeners()



from subprocess import Popen, PIPE, STDOUT


def exe_command(command):
    """
    执行 shell 命令并实时打印输出
    :param command: shell 命令
    :return: process, exitcode
    """
    print(command)
    process = Popen(command, stdout=PIPE, stderr=STDOUT, shell=True)
    with process.stdout:
        for line in iter(process.stdout.readline, b''):
            print(line.decode().strip())
    exitcode = process.wait()
    return process, exitcode



import platform
import subprocess
import glob
import re
import packaging
from packaging import version
import requests
import zipfile
import os
import stat

"""
自动下载chromedriver
"""


##simple script to test capturing version info from multi OS types and versions

# capture platform information
def getOSInfo():
    osPlatform = (platform.platform())
    # mac version capture
    if 'macos' in osPlatform.lower():
        osType = 'mac'
        chromeVersion = list(set(re.findall("(?:[0-9]*\.){3}[0-9]*", subprocess.run(
            ["/Applications/Google Chrome.app/Contents/MacOS/Google Chrome", "--version"], capture_output=True,
            text=True).stdout)))
        chromeVersion = chromeVersion[0]
        # is the processor intel or m1?
        if platform.processor() == 'i386':
            driverType = 'mac64'
        if platform.processor() == 'arm':
            driverType = 'mac64_m1'
    # windows version capture
    elif 'windows' in osPlatform.lower():
        print(osPlatform)
        osType = 'win'
        # handle multiple install directories
        chromeDir = glob.glob("C:\\Program Files (x86)\\Google\\Chrome\\Application\\**\\", recursive=True)
        if not chromeDir:
            chromeDir = glob.glob("C:\\Program Files\\Google\\Chrome\\Application\\**\\", recursive=True)
        chromeDirStr = ''.join(chromeDir)
        parsedVers = []
        for i in list(set(re.findall("(?:[0-9]*\.){3}[0-9]*", chromeDirStr))):
            parsedVers.append(version.parse(i))
        chromeVersion = str(max(parsedVers))
        driverType = 'win32'
    # linux version capture
    elif 'linux' in osPlatform.lower():
        osType = 'linux'
        chromeVersion = subprocess.run(["google-chrome", "--version"])
        driverType = 'linux64'
    print("检测到当前chrome版本为{}:{}".format(driverType, chromeVersion))
    return chromeVersion, driverType


def verifyDirectoryExists(path):
    if not os.path.isdir(path):
        print(f"{path} doesn't exist...creating")
        os.makedirs(path)
    return os.path.isdir(path)


def removeZip(path):
    if os.path.isfile(path):
        os.remove(path)
    return os.path.isfile(path)


def getDriver(outputPath=None, deleteZip=True):
    if outputPath:
        if outputPath[-1] == '/' or outputPath[-1] == '\\':
            outputPath = outputPath[:-1]
        if '~' in outputPath:
            outputPath = os.path.expanduser(outputPath)
        verifyDirectoryExists(outputPath)
    chromeVersion, driverType = getOSInfo()
    print("正在下载chromedriver")
    # chrome driver version must match major version of chrome (1st Octet)
    majorVersion = chromeVersion.split('.')[0]
    chromeDriverVersion = requests.get(
        f"https://chromedriver.storage.googleapis.com/LATEST_RELEASE_{chromeVersion.split('.')[0]}").text
    response = requests.get(
        f'https://chromedriver.storage.googleapis.com/{chromeDriverVersion}/chromedriver_{driverType}.zip')
    totalbits = 0
    if response.status_code == 200:
        if not outputPath:
            outputPath = '.'
        zipLocation = f'{outputPath}/chromedriver.zip'
        with open(zipLocation, 'wb') as f:
            for chunk in response.iter_content(chunk_size=1024):
                if chunk:
                    totalbits += 1024
                    print("Downloaded", totalbits * 1025, "KB...")
                    f.write(chunk)
            f.close()
    with zipfile.ZipFile(zipLocation, 'r') as zip_ref:
        zip_ref.extractall(f'{outputPath}')
    os.chmod(zipLocation, 0o755)
    if driverType == 'win32':
        os.chmod(f'{outputPath}/chromedriver.exe', 0o755)
    else:
        os.chmod(f'{outputPath}/chromedriver', 0o755)
    if deleteZip:
        print("Chromedriver extracted, removing zip file")
        removeZip(zipLocation)
    else:
        print("Chromedriver Extracted")
    return (os.listdir(outputPath))


def main():
    getDriver()


if __name__ == "__main__":
    main()



# Copyright 2014-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

"""A fake SSLContext implementation."""

import ssl as _ssl
import sys as _sys

# PROTOCOL_TLS_CLIENT is Python 3.6+
PROTOCOL_SSLv23 = getattr(_ssl, "PROTOCOL_TLS_CLIENT", _ssl.PROTOCOL_SSLv23)
OP_NO_SSLv2 = getattr(_ssl, "OP_NO_SSLv2", 0)
OP_NO_SSLv3 = getattr(_ssl, "OP_NO_SSLv3", 0)
OP_NO_COMPRESSION = getattr(_ssl, "OP_NO_COMPRESSION", 0)
# Python 3.7+, OpenSSL 1.1.0h+
OP_NO_RENEGOTIATION = getattr(_ssl, "OP_NO_RENEGOTIATION", 0)

HAS_SNI = getattr(_ssl, "HAS_SNI", False)
IS_PYOPENSSL = False

# Base Exception class
SSLError = _ssl.SSLError

from ssl import SSLContext
if hasattr(_ssl, "VERIFY_CRL_CHECK_LEAF"):
    from ssl import VERIFY_CRL_CHECK_LEAF
# Python 3.7 uses OpenSSL's hostname matching implementation
# making it the obvious version to start using SSLConext.check_hostname.
# Python 3.6 might have been a good version, but it suffers
# from https://bugs.python.org/issue32185.
CHECK_HOSTNAME_SAFE = _sys.version_info[:2] >= (3, 7)



# Copyright 2014-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

"""Internal class to monitor a topology of one or more servers."""

import os
import queue
import random
import threading
import time
import warnings
import weakref

from pymongo import (common,
                     helpers,
                     periodic_executor)
from pymongo.client_session import _ServerSessionPool
from pymongo.errors import (ConnectionFailure,
                            ConfigurationError,
                            NetworkTimeout,
                            NotPrimaryError,
                            OperationFailure,
                            PyMongoError,
                            ServerSelectionTimeoutError,
                            WriteError,
                            InvalidOperation)
from pymongo.hello import Hello
from pymongo.monitor import SrvMonitor
from pymongo.pool import PoolOptions
from pymongo.server import Server
from pymongo.server_description import ServerDescription
from pymongo.server_selectors import (any_server_selector,
                                      arbiter_server_selector,
                                      secondary_server_selector,
                                      readable_server_selector,
                                      writable_server_selector,
                                      Selection)
from pymongo.topology_description import (updated_topology_description,
                                          _updated_topology_description_srv_polling,
                                          TopologyDescription,
                                          SRV_POLLING_TOPOLOGIES, TOPOLOGY_TYPE)


def process_events_queue(queue_ref):
    q = queue_ref()
    if not q:
        return False  # Cancel PeriodicExecutor.

    while True:
        try:
            event = q.get_nowait()
        except queue.Empty:
            break
        else:
            fn, args = event
            fn(*args)

    return True  # Continue PeriodicExecutor.


class Topology(object):
    """Monitor a topology of one or more servers."""
    def __init__(self, topology_settings):
        self._topology_id = topology_settings._topology_id
        self._listeners = topology_settings._pool_options._event_listeners
        pub = self._listeners is not None
        self._publish_server = pub and self._listeners.enabled_for_server
        self._publish_tp = pub and self._listeners.enabled_for_topology

        # Create events queue if there are publishers.
        self._events = None
        self.__events_executor = None

        if self._publish_server or self._publish_tp:
            self._events = queue.Queue(maxsize=100)

        if self._publish_tp:
            self._events.put((self._listeners.publish_topology_opened,
                             (self._topology_id,)))
        self._settings = topology_settings
        topology_description = TopologyDescription(
            topology_settings.get_topology_type(),
            topology_settings.get_server_descriptions(),
            topology_settings.replica_set_name,
            None,
            None,
            topology_settings)

        self._description = topology_description
        if self._publish_tp:
            initial_td = TopologyDescription(TOPOLOGY_TYPE.Unknown, {}, None,
                                             None, None, self._settings)
            self._events.put((
                self._listeners.publish_topology_description_changed,
                (initial_td, self._description, self._topology_id)))

        for seed in topology_settings.seeds:
            if self._publish_server:
                self._events.put((self._listeners.publish_server_opened,
                                 (seed, self._topology_id)))

        # Store the seed list to help diagnose errors in _error_message().
        self._seed_addresses = list(topology_description.server_descriptions())
        self._opened = False
        self._closed = False
        self._lock = threading.Lock()
        self._condition = self._settings.condition_class(self._lock)
        self._servers = {}
        self._pid = None
        self._max_cluster_time = None
        self._session_pool = _ServerSessionPool()

        if self._publish_server or self._publish_tp:
            def target():
                return process_events_queue(weak)

            executor = periodic_executor.PeriodicExecutor(
                interval=common.EVENTS_QUEUE_FREQUENCY,
                min_interval=common.MIN_HEARTBEAT_INTERVAL,
                target=target,
                name="pymongo_events_thread")

            # We strongly reference the executor and it weakly references
            # the queue via this closure. When the topology is freed, stop
            # the executor soon.
            weak = weakref.ref(self._events, executor.close)
            self.__events_executor = executor
            executor.open()

        self._srv_monitor = None
        if (self._settings.fqdn is not None and
                not self._settings.load_balanced):
            self._srv_monitor = SrvMonitor(self, self._settings)

    def open(self):
        """Start monitoring, or restart after a fork.

        No effect if called multiple times.

        .. warning:: Topology is shared among multiple threads and is protected
          by mutual exclusion. Using Topology from a process other than the one
          that initialized it will emit a warning and may result in deadlock. To
          prevent this from happening, MongoClient must be created after any
          forking.

        """
        if self._pid is None:
            self._pid = os.getpid()
        else:
            if os.getpid() != self._pid:
                warnings.warn(
                    "MongoClient opened before fork. Create MongoClient only "
                    "after forking. See PyMongo's documentation for details: "
                    "https://pymongo.readthedocs.io/en/stable/faq.html#"
                    "is-pymongo-fork-safe")
                with self._lock:
                    # Reset the session pool to avoid duplicate sessions in
                    # the child process.
                    self._session_pool.reset()

        with self._lock:
            self._ensure_opened()

    def select_servers(self,
                       selector,
                       server_selection_timeout=None,
                       address=None):
        """Return a list of Servers matching selector, or time out.

        :Parameters:
          - `selector`: function that takes a list of Servers and returns
            a subset of them.
          - `server_selection_timeout` (optional): maximum seconds to wait.
            If not provided, the default value common.SERVER_SELECTION_TIMEOUT
            is used.
          - `address`: optional server address to select.

        Calls self.open() if needed.

        Raises exc:`ServerSelectionTimeoutError` after
        `server_selection_timeout` if no matching servers are found.
        """
        if server_selection_timeout is None:
            server_timeout = self._settings.server_selection_timeout
        else:
            server_timeout = server_selection_timeout

        with self._lock:
            server_descriptions = self._select_servers_loop(
                selector, server_timeout, address)

            return [self.get_server_by_address(sd.address)
                    for sd in server_descriptions]

    def _select_servers_loop(self, selector, timeout, address):
        """select_servers() guts. Hold the lock when calling this."""
        now = time.monotonic()
        end_time = now + timeout
        server_descriptions = self._description.apply_selector(
            selector, address, custom_selector=self._settings.server_selector)

        while not server_descriptions:
            # No suitable servers.
            if timeout == 0 or now > end_time:
                raise ServerSelectionTimeoutError(
                    "%s, Timeout: %ss, Topology Description: %r" %
                    (self._error_message(selector), timeout, self.description))

            self._ensure_opened()
            self._request_check_all()

            # Release the lock and wait for the topology description to
            # change, or for a timeout. We won't miss any changes that
            # came after our most recent apply_selector call, since we've
            # held the lock until now.
            self._condition.wait(common.MIN_HEARTBEAT_INTERVAL)
            self._description.check_compatible()
            now = time.monotonic()
            server_descriptions = self._description.apply_selector(
                selector, address,
                custom_selector=self._settings.server_selector)

        self._description.check_compatible()
        return server_descriptions

    def select_server(self,
                      selector,
                      server_selection_timeout=None,
                      address=None):
        """Like select_servers, but choose a random server if several match."""
        servers = self.select_servers(
            selector, server_selection_timeout, address)
        if len(servers) == 1:
            return servers[0]
        server1, server2 = random.sample(servers, 2)
        if server1.pool.operation_count <= server2.pool.operation_count:
            return server1
        else:
            return server2

    def select_server_by_address(self, address,
                                 server_selection_timeout=None):
        """Return a Server for "address", reconnecting if necessary.

        If the server's type is not known, request an immediate check of all
        servers. Time out after "server_selection_timeout" if the server
        cannot be reached.

        :Parameters:
          - `address`: A (host, port) pair.
          - `server_selection_timeout` (optional): maximum seconds to wait.
            If not provided, the default value
            common.SERVER_SELECTION_TIMEOUT is used.

        Calls self.open() if needed.

        Raises exc:`ServerSelectionTimeoutError` after
        `server_selection_timeout` if no matching servers are found.
        """
        return self.select_server(any_server_selector,
                                  server_selection_timeout,
                                  address)

    def _process_change(self, server_description, reset_pool=False):
        """Process a new ServerDescription on an opened topology.

        Hold the lock when calling this.
        """
        td_old = self._description
        sd_old = td_old._server_descriptions[server_description.address]
        if _is_stale_server_description(sd_old, server_description):
            # This is a stale hello response. Ignore it.
            return

        new_td = updated_topology_description(
            self._description, server_description)
        # CMAP: Ensure the pool is "ready" when the server is selectable.
        if (server_description.is_readable
                or (server_description.is_server_type_known and
                    new_td.topology_type == TOPOLOGY_TYPE.Single)):
            server = self._servers.get(server_description.address)
            if server:
                server.pool.ready()

        suppress_event = ((self._publish_server or self._publish_tp)
                          and sd_old == server_description)
        if self._publish_server and not suppress_event:
            self._events.put((
                self._listeners.publish_server_description_changed,
                (sd_old, server_description,
                 server_description.address, self._topology_id)))

        self._description = new_td
        self._update_servers()
        self._receive_cluster_time_no_lock(server_description.cluster_time)

        if self._publish_tp and not suppress_event:
            self._events.put((
                self._listeners.publish_topology_description_changed,
                (td_old, self._description, self._topology_id)))

        # Shutdown SRV polling for unsupported cluster types.
        # This is only applicable if the old topology was Unknown, and the
        # new one is something other than Unknown or Sharded.
        if self._srv_monitor and (td_old.topology_type == TOPOLOGY_TYPE.Unknown
                                  and self._description.topology_type not in
                                  SRV_POLLING_TOPOLOGIES):
            self._srv_monitor.close()

        # Clear the pool from a failed heartbeat.
        if reset_pool:
            server = self._servers.get(server_description.address)
            if server:
                server.pool.reset()

        # Wake waiters in select_servers().
        self._condition.notify_all()

    def on_change(self, server_description, reset_pool=False):
        """Process a new ServerDescription after an hello call completes."""
        # We do no I/O holding the lock.
        with self._lock:
            # Monitors may continue working on hello calls for some time
            # after a call to Topology.close, so this method may be called at
            # any time. Ensure the topology is open before processing the
            # change.
            # Any monitored server was definitely in the topology description
            # once. Check if it's still in the description or if some state-
            # change removed it. E.g., we got a host list from the primary
            # that didn't include this server.
            if (self._opened and
                    self._description.has_server(server_description.address)):
                self._process_change(server_description, reset_pool)

    def _process_srv_update(self, seedlist):
        """Process a new seedlist on an opened topology.
        Hold the lock when calling this.
        """
        td_old = self._description
        self._description = _updated_topology_description_srv_polling(
            self._description, seedlist)

        self._update_servers()

        if self._publish_tp:
            self._events.put((
                self._listeners.publish_topology_description_changed,
                (td_old, self._description, self._topology_id)))

    def on_srv_update(self, seedlist):
        """Process a new list of nodes obtained from scanning SRV records."""
        # We do no I/O holding the lock.
        with self._lock:
            if self._opened:
                self._process_srv_update(seedlist)

    def get_server_by_address(self, address):
        """Get a Server or None.

        Returns the current version of the server immediately, even if it's
        Unknown or absent from the topology. Only use this in unittests.
        In driver code, use select_server_by_address, since then you're
        assured a recent view of the server's type and wire protocol version.
        """
        return self._servers.get(address)

    def has_server(self, address):
        return address in self._servers

    def get_primary(self):
        """Return primary's address or None."""
        # Implemented here in Topology instead of MongoClient, so it can lock.
        with self._lock:
            topology_type = self._description.topology_type
            if topology_type != TOPOLOGY_TYPE.ReplicaSetWithPrimary:
                return None

            return writable_server_selector(self._new_selection())[0].address

    def _get_replica_set_members(self, selector):
        """Return set of replica set member addresses."""
        # Implemented here in Topology instead of MongoClient, so it can lock.
        with self._lock:
            topology_type = self._description.topology_type
            if topology_type not in (TOPOLOGY_TYPE.ReplicaSetWithPrimary,
                                     TOPOLOGY_TYPE.ReplicaSetNoPrimary):
                return set()

            return set([sd.address for sd in selector(self._new_selection())])

    def get_secondaries(self):
        """Return set of secondary addresses."""
        return self._get_replica_set_members(secondary_server_selector)

    def get_arbiters(self):
        """Return set of arbiter addresses."""
        return self._get_replica_set_members(arbiter_server_selector)

    def max_cluster_time(self):
        """Return a document, the highest seen $clusterTime."""
        return self._max_cluster_time

    def _receive_cluster_time_no_lock(self, cluster_time):
        # Driver Sessions Spec: "Whenever a driver receives a cluster time from
        # a server it MUST compare it to the current highest seen cluster time
        # for the deployment. If the new cluster time is higher than the
        # highest seen cluster time it MUST become the new highest seen cluster
        # time. Two cluster times are compared using only the BsonTimestamp
        # value of the clusterTime embedded field."
        if cluster_time:
            # ">" uses bson.timestamp.Timestamp's comparison operator.
            if (not self._max_cluster_time
                or cluster_time['clusterTime'] >
                    self._max_cluster_time['clusterTime']):
                self._max_cluster_time = cluster_time

    def receive_cluster_time(self, cluster_time):
        with self._lock:
            self._receive_cluster_time_no_lock(cluster_time)

    def request_check_all(self, wait_time=5):
        """Wake all monitors, wait for at least one to check its server."""
        with self._lock:
            self._request_check_all()
            self._condition.wait(wait_time)

    def data_bearing_servers(self):
        """Return a list of all data-bearing servers.

        This includes any server that might be selected for an operation.
        """
        if self._description.topology_type == TOPOLOGY_TYPE.Single:
            return self._description.known_servers
        return self._description.readable_servers

    def update_pool(self, all_credentials):
        # Remove any stale sockets and add new sockets if pool is too small.
        servers = []
        with self._lock:
            # Only update pools for data-bearing servers.
            for sd in self.data_bearing_servers():
                server = self._servers[sd.address]
                servers.append((server,
                                server.pool.gen.get_overall()))

        for server, generation in servers:
            try:
                server.pool.remove_stale_sockets(generation, all_credentials)
            except PyMongoError as exc:
                ctx = _ErrorContext(exc, 0, generation, False, None)
                self.handle_error(server.description.address, ctx)
                raise

    def close(self):
        """Clear pools and terminate monitors. Topology does not reopen on
        demand. Any further operations will raise
        :exc:`~.errors.InvalidOperation`. """
        with self._lock:
            for server in self._servers.values():
                server.close()

            # Mark all servers Unknown.
            self._description = self._description.reset()
            for address, sd in self._description.server_descriptions().items():
                if address in self._servers:
                    self._servers[address].description = sd

            # Stop SRV polling thread.
            if self._srv_monitor:
                self._srv_monitor.close()

            self._opened = False
            self._closed = True

        # Publish only after releasing the lock.
        if self._publish_tp:
            self._events.put((self._listeners.publish_topology_closed,
                              (self._topology_id,)))
        if self._publish_server or self._publish_tp:
            self.__events_executor.close()

    @property
    def description(self):
        return self._description

    def pop_all_sessions(self):
        """Pop all session ids from the pool."""
        with self._lock:
            return self._session_pool.pop_all()

    def _check_session_support(self):
        """Internal check for session support on non-load balanced clusters."""
        session_timeout = self._description.logical_session_timeout_minutes
        if session_timeout is None:
            # Maybe we need an initial scan? Can raise ServerSelectionError.
            if self._description.topology_type == TOPOLOGY_TYPE.Single:
                if not self._description.has_known_servers:
                    self._select_servers_loop(
                        any_server_selector,
                        self._settings.server_selection_timeout,
                        None)
            elif not self._description.readable_servers:
                self._select_servers_loop(
                    readable_server_selector,
                    self._settings.server_selection_timeout,
                    None)

            session_timeout = self._description.logical_session_timeout_minutes
            if session_timeout is None:
                raise ConfigurationError(
                    "Sessions are not supported by this MongoDB deployment")
        return session_timeout

    def get_server_session(self):
        """Start or resume a server session, or raise ConfigurationError."""
        with self._lock:
            # Sessions are always supported in load balanced mode.
            if not self._settings.load_balanced:
                session_timeout = self._check_session_support()
            else:
                # Sessions never time out in load balanced mode.
                session_timeout = float('inf')
            return self._session_pool.get_server_session(session_timeout)

    def return_server_session(self, server_session, lock):
        if lock:
            with self._lock:
                self._session_pool.return_server_session(
                    server_session,
                    self._description.logical_session_timeout_minutes)
        else:
            # Called from a __del__ method, can't use a lock.
            self._session_pool.return_server_session_no_lock(server_session)

    def _new_selection(self):
        """A Selection object, initially including all known servers.

        Hold the lock when calling this.
        """
        return Selection.from_topology_description(self._description)

    def _ensure_opened(self):
        """Start monitors, or restart after a fork.

        Hold the lock when calling this.
        """
        if self._closed:
            raise InvalidOperation("Cannot use MongoClient after close")

        if not self._opened:
            self._opened = True
            self._update_servers()

            # Start or restart the events publishing thread.
            if self._publish_tp or self._publish_server:
                self.__events_executor.open()

            # Start the SRV polling thread.
            if self._srv_monitor and (self.description.topology_type in
                                      SRV_POLLING_TOPOLOGIES):
                self._srv_monitor.open()

            if self._settings.load_balanced:
                # Emit initial SDAM events for load balancer mode.
                self._process_change(ServerDescription(
                    self._seed_addresses[0],
                    Hello({'ok': 1, 'serviceId': self._topology_id,
                              'maxWireVersion': 13})))

        # Ensure that the monitors are open.
        for server in self._servers.values():
            server.open()

    def _is_stale_error(self, address, err_ctx):
        server = self._servers.get(address)
        if server is None:
            # Another thread removed this server from the topology.
            return True

        if server._pool.stale_generation(
                err_ctx.sock_generation, err_ctx.service_id):
            # This is an outdated error from a previous pool version.
            return True

        # topologyVersion check, ignore error when cur_tv >= error_tv:
        cur_tv = server.description.topology_version
        error = err_ctx.error
        error_tv = None
        if error and hasattr(error, 'details'):
            if isinstance(error.details, dict):
                error_tv = error.details.get('topologyVersion')

        return _is_stale_error_topology_version(cur_tv, error_tv)

    def _handle_error(self, address, err_ctx):
        if self._is_stale_error(address, err_ctx):
            return

        server = self._servers[address]
        error = err_ctx.error
        exc_type = type(error)
        service_id = err_ctx.service_id
        if (issubclass(exc_type, NetworkTimeout) and
                err_ctx.completed_handshake):
            # The socket has been closed. Don't reset the server.
            # Server Discovery And Monitoring Spec: "When an application
            # operation fails because of any network error besides a socket
            # timeout...."
            return
        elif issubclass(exc_type, WriteError):
            # Ignore writeErrors.
            return
        elif issubclass(exc_type, (NotPrimaryError, OperationFailure)):
            # As per the SDAM spec if:
            #   - the server sees a "not primary" error, and
            #   - the server is not shutting down, and
            #   - the server version is >= 4.2, then
            # we keep the existing connection pool, but mark the server type
            # as Unknown and request an immediate check of the server.
            # Otherwise, we clear the connection pool, mark the server as
            # Unknown and request an immediate check of the server.
            if hasattr(error, 'code'):
                err_code = error.code
            else:
                # Default error code if one does not exist.
                default = 10107 if isinstance(error, NotPrimaryError) else None
                err_code = error.details.get('code', default)
            if err_code in helpers._NOT_PRIMARY_CODES:
                is_shutting_down = err_code in helpers._SHUTDOWN_CODES
                # Mark server Unknown, clear the pool, and request check.
                if not self._settings.load_balanced:
                    self._process_change(ServerDescription(address, error=error))
                if is_shutting_down or (err_ctx.max_wire_version <= 7):
                    # Clear the pool.
                    server.reset(service_id)
                server.request_check()
            elif not err_ctx.completed_handshake:
                # Unknown command error during the connection handshake.
                if not self._settings.load_balanced:
                    self._process_change(ServerDescription(address, error=error))
                # Clear the pool.
                server.reset(service_id)
        elif issubclass(exc_type, ConnectionFailure):
            # "Client MUST replace the server's description with type Unknown
            # ... MUST NOT request an immediate check of the server."
            if not self._settings.load_balanced:
                self._process_change(ServerDescription(address, error=error))
            # Clear the pool.
            server.reset(service_id)
            # "When a client marks a server Unknown from `Network error when
            # reading or writing`_, clients MUST cancel the hello check on
            # that server and close the current monitoring connection."
            server._monitor.cancel_check()

    def handle_error(self, address, err_ctx):
        """Handle an application error.

        May reset the server to Unknown, clear the pool, and request an
        immediate check depending on the error and the context.
        """
        with self._lock:
            self._handle_error(address, err_ctx)

    def _request_check_all(self):
        """Wake all monitors. Hold the lock when calling this."""
        for server in self._servers.values():
            server.request_check()

    def _update_servers(self):
        """Sync our Servers from TopologyDescription.server_descriptions.

        Hold the lock while calling this.
        """
        for address, sd in self._description.server_descriptions().items():
            if address not in self._servers:
                monitor = self._settings.monitor_class(
                    server_description=sd,
                    topology=self,
                    pool=self._create_pool_for_monitor(address),
                    topology_settings=self._settings)

                weak = None
                if self._publish_server:
                    weak = weakref.ref(self._events)
                server = Server(
                    server_description=sd,
                    pool=self._create_pool_for_server(address),
                    monitor=monitor,
                    topology_id=self._topology_id,
                    listeners=self._listeners,
                    events=weak)

                self._servers[address] = server
                server.open()
            else:
                # Cache old is_writable value.
                was_writable = self._servers[address].description.is_writable
                # Update server description.
                self._servers[address].description = sd
                # Update is_writable value of the pool, if it changed.
                if was_writable != sd.is_writable:
                    self._servers[address].pool.update_is_writable(
                        sd.is_writable)

        for address, server in list(self._servers.items()):
            if not self._description.has_server(address):
                server.close()
                self._servers.pop(address)

    def _create_pool_for_server(self, address):
        return self._settings.pool_class(address, self._settings.pool_options)

    def _create_pool_for_monitor(self, address):
        options = self._settings.pool_options

        # According to the Server Discovery And Monitoring Spec, monitors use
        # connect_timeout for both connect_timeout and socket_timeout. The
        # pool only has one socket so maxPoolSize and so on aren't needed.
        monitor_pool_options = PoolOptions(
            connect_timeout=options.connect_timeout,
            socket_timeout=options.connect_timeout,
            ssl_context=options._ssl_context,
            tls_allow_invalid_hostnames=options.tls_allow_invalid_hostnames,
            event_listeners=options._event_listeners,
            appname=options.appname,
            driver=options.driver,
            pause_enabled=False,
            server_api=options.server_api,
        )

        return self._settings.pool_class(address, monitor_pool_options,
                                         handshake=False)

    def _error_message(self, selector):
        """Format an error message if server selection fails.

        Hold the lock when calling this.
        """
        is_replica_set = self._description.topology_type in (
            TOPOLOGY_TYPE.ReplicaSetWithPrimary,
            TOPOLOGY_TYPE.ReplicaSetNoPrimary)

        if is_replica_set:
            server_plural = 'replica set members'
        elif self._description.topology_type == TOPOLOGY_TYPE.Sharded:
            server_plural = 'mongoses'
        else:
            server_plural = 'servers'

        if self._description.known_servers:
            # We've connected, but no servers match the selector.
            if selector is writable_server_selector:
                if is_replica_set:
                    return 'No primary available for writes'
                else:
                    return 'No %s available for writes' % server_plural
            else:
                return 'No %s match selector "%s"' % (server_plural, selector)
        else:
            addresses = list(self._description.server_descriptions())
            servers = list(self._description.server_descriptions().values())
            if not servers:
                if is_replica_set:
                    # We removed all servers because of the wrong setName?
                    return 'No %s available for replica set name "%s"' % (
                        server_plural, self._settings.replica_set_name)
                else:
                    return 'No %s available' % server_plural

            # 1 or more servers, all Unknown. Are they unknown for one reason?
            error = servers[0].error
            same = all(server.error == error for server in servers[1:])
            if same:
                if error is None:
                    # We're still discovering.
                    return 'No %s found yet' % server_plural

                if (is_replica_set and not
                        set(addresses).intersection(self._seed_addresses)):
                    # We replaced our seeds with new hosts but can't reach any.
                    return (
                        'Could not reach any servers in %s. Replica set is'
                        ' configured with internal hostnames or IPs?' %
                        addresses)

                return str(error)
            else:
                return ','.join(str(server.error) for server in servers
                                if server.error)

    def __repr__(self):
        msg = ''
        if not self._opened:
            msg = 'CLOSED '
        return '<%s %s%r>' % (self.__class__.__name__, msg, self._description)

    def eq_props(self):
        """The properties to use for MongoClient/Topology equality checks."""
        ts = self._settings
        return (tuple(sorted(ts.seeds)), ts.replica_set_name, ts.fqdn,
                ts.srv_service_name)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.eq_props() == other.eq_props()
        return NotImplemented

    def __hash__(self):
        return hash(self.eq_props())


class _ErrorContext(object):
    """An error with context for SDAM error handling."""
    def __init__(self, error, max_wire_version, sock_generation,
                 completed_handshake, service_id):
        self.error = error
        self.max_wire_version = max_wire_version
        self.sock_generation = sock_generation
        self.completed_handshake = completed_handshake
        self.service_id = service_id


def _is_stale_error_topology_version(current_tv, error_tv):
    """Return True if the error's topologyVersion is <= current."""
    if current_tv is None or error_tv is None:
        return False
    if current_tv['processId'] != error_tv['processId']:
        return False
    return current_tv['counter'] >= error_tv['counter']


def _is_stale_server_description(current_sd, new_sd):
    """Return True if the new topologyVersion is < current."""
    current_tv, new_tv = current_sd.topology_version, new_sd.topology_version
    if current_tv is None or new_tv is None:
        return False
    if current_tv['processId'] != new_tv['processId']:
        return False
    return current_tv['counter'] > new_tv['counter']



# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import NoReturn
from selenium.webdriver.common.options import BaseOptions
from selenium.webdriver.common.service import Service
from selenium.webdriver.edge.options import Options as EdgeOptions
from selenium.webdriver.chrome.options import Options as ChromeOptions
import warnings

from selenium.webdriver.chromium.remote_connection import ChromiumRemoteConnection
from selenium.webdriver.remote.webdriver import WebDriver as RemoteWebDriver

DEFAULT_PORT = 0
DEFAULT_SERVICE_LOG_PATH = None
DEFAULT_KEEP_ALIVE = None


class ChromiumDriver(RemoteWebDriver):
    """
    Controls the WebDriver instance of ChromiumDriver and allows you to drive the browser.
    """

    def __init__(self, browser_name, vendor_prefix,
                 port=DEFAULT_PORT, options: BaseOptions = None, service_args=None,
                 desired_capabilities=None, service_log_path=DEFAULT_SERVICE_LOG_PATH,
                 service: Service = None, keep_alive=DEFAULT_KEEP_ALIVE):
        """
        Creates a new WebDriver instance of the ChromiumDriver.
        Starts the service and then creates new WebDriver instance of ChromiumDriver.

        :Args:
         - browser_name - Browser name used when matching capabilities.
         - vendor_prefix - Company prefix to apply to vendor-specific WebDriver extension commands.
         - port - Deprecated: port you would like the service to run, if left as 0, a free port will be found.
         - options - this takes an instance of ChromiumOptions
         - service_args - Deprecated: List of args to pass to the driver service
         - desired_capabilities - Deprecated: Dictionary object with non-browser specific
           capabilities only, such as "proxy" or "loggingPref".
         - service_log_path - Deprecated: Where to log information from the driver.
         - keep_alive - Deprecated: Whether to configure ChromiumRemoteConnection to use HTTP keep-alive.
        """
        if desired_capabilities:
            warnings.warn('desired_capabilities has been deprecated, please pass in a Service object',
                          DeprecationWarning, stacklevel=2)
        if port != DEFAULT_PORT:
            warnings.warn('port has been deprecated, please pass in a Service object',
                          DeprecationWarning, stacklevel=2)
        self.port = port
        if service_log_path != DEFAULT_SERVICE_LOG_PATH:
            warnings.warn('service_log_path has been deprecated, please pass in a Service object',
                          DeprecationWarning, stacklevel=2)
        if keep_alive != DEFAULT_KEEP_ALIVE and type(self) == __class__:
            warnings.warn('keep_alive has been deprecated, please pass in a Service object',
                          DeprecationWarning, stacklevel=2)
        else:
            keep_alive = True

        self.vendor_prefix = vendor_prefix

        _ignore_proxy = None
        if not options:
            options = self.create_options()

        if desired_capabilities:
            for key, value in desired_capabilities.items():
                options.set_capability(key, value)

        if options._ignore_local_proxy:
            _ignore_proxy = options._ignore_local_proxy

        if not service:
            raise AttributeError('service cannot be None')

        self.service = service
        self.service.start()

        try:
            RemoteWebDriver.__init__(
                self,
                command_executor=ChromiumRemoteConnection(
                    remote_server_addr=self.service.service_url,
                    browser_name=browser_name, vendor_prefix=vendor_prefix,
                    keep_alive=keep_alive, ignore_proxy=_ignore_proxy),
                options=options)
        except Exception:
            self.quit()
            raise
        self._is_remote = False

    def launch_app(self, id):
        """Launches Chromium app specified by id."""
        return self.execute("launchApp", {'id': id})

    def get_network_conditions(self):
        """
        Gets Chromium network emulation settings.

        :Returns:
            A dict. For example:
            {'latency': 4, 'download_throughput': 2, 'upload_throughput': 2,
            'offline': False}
        """
        return self.execute("getNetworkConditions")['value']

    def set_network_conditions(self, **network_conditions) -> NoReturn:
        """
        Sets Chromium network emulation settings.

        :Args:
         - network_conditions: A dict with conditions specification.

        :Usage:
            ::

                driver.set_network_conditions(
                    offline=False,
                    latency=5,  # additional latency (ms)
                    download_throughput=500 * 1024,  # maximal throughput
                    upload_throughput=500 * 1024)  # maximal throughput

            Note: 'throughput' can be used to set both (for download and upload).
        """
        self.execute("setNetworkConditions", {
            'network_conditions': network_conditions
        })

    def delete_network_conditions(self) -> NoReturn:
        """
        Resets Chromium network emulation settings.
        """
        self.execute("deleteNetworkConditions")

    def set_permissions(self, name: str, value: str) -> NoReturn:
        """
        Sets Applicable Permission.

        :Args:
         - name: The item to set the permission on.
         - value: The value to set on the item

        :Usage:
            ::
                driver.set_permissions('clipboard-read', 'denied')
        """
        self.execute("setPermissions", {'descriptor': {'name': name}, 'state': value})

    def execute_cdp_cmd(self, cmd: str, cmd_args: dict):
        """
        Execute Chrome Devtools Protocol command and get returned result
        The command and command args should follow chrome devtools protocol domains/commands, refer to link
        https://chromedevtools.github.io/devtools-protocol/

        :Args:
         - cmd: A str, command name
         - cmd_args: A dict, command args. empty dict {} if there is no command args
        :Usage:
            ::
                driver.execute_cdp_cmd('Network.getResponseBody', {'requestId': requestId})
        :Returns:
            A dict, empty dict {} if there is no result to return.
            For example to getResponseBody:
            {'base64Encoded': False, 'body': 'response body string'}
        """
        return self.execute("executeCdpCommand", {'cmd': cmd, 'params': cmd_args})['value']

    def get_sinks(self) -> list:
        """
        :Returns: A list of sinks available for Cast.
        """
        return self.execute('getSinks')['value']

    def get_issue_message(self):
        """
        :Returns: An error message when there is any issue in a Cast session.
        """
        return self.execute('getIssueMessage')['value']

    def set_sink_to_use(self, sink_name: str) -> str:
        """
        Sets a specific sink, using its name, as a Cast session receiver target.

        :Args:
         - sink_name: Name of the sink to use as the target.
        """
        return self.execute('setSinkToUse', {'sinkName': sink_name})

    def start_tab_mirroring(self, sink_name: str) -> str:
        """
        Starts a tab mirroring session on a specific receiver target.

        :Args:
         - sink_name: Name of the sink to use as the target.
        """
        return self.execute('startTabMirroring', {'sinkName': sink_name})

    def stop_casting(self, sink_name: str) -> str:
        """
        Stops the existing Cast session on a specific receiver target.

        :Args:
         - sink_name: Name of the sink to stop the Cast session.
        """
        return self.execute('stopCasting', {'sinkName': sink_name})

    def quit(self) -> NoReturn:
        """
        Closes the browser and shuts down the ChromiumDriver executable
        that is started when starting the ChromiumDriver
        """
        try:
            RemoteWebDriver.quit(self)
        except Exception:
            # We don't care about the message because something probably has gone wrong
            pass
        finally:
            self.service.stop()

    def create_options(self) -> BaseOptions:
        if self.vendor_prefix == "ms":
            return EdgeOptions()
        else:
            return ChromeOptions()



import os
from random_user_agent.user_agent import UserAgent
from random_user_agent.params import SoftwareName, OperatingSystem
from similib.conf import SIMILIB_BASE_DIR

with open(os.path.join(SIMILIB_BASE_DIR, "file", "stealth.min.js"), "r") as f:
    stealth_js = f.read()


def hide_security(bro):
    bro.execute_cdp_cmd("Page.addScriptToEvaluateOnNewDocument", {
        "source": stealth_js
    })


software_names = [SoftwareName.CHROME.value]
operating_systems = [OperatingSystem.WINDOWS.value, OperatingSystem.LINUX.value]

user_agent_rotator = UserAgent(software_names=software_names, operating_systems=operating_systems, limit=100)


def get_user_agents():
    return user_agent_rotator.get_user_agents()


def get_random_user_agent():
    return user_agent_rotator.get_random_user_agent()





from enum import Enum


class SoftwareName(Enum):
    ALERTSITE_MONITORING_BOT = 'alertsite-monitoring-bot'
    ANDROID = 'android-browser'
    AWESOMIUM = 'awesomium'
    BLACKBERRY = 'blackberry-browser'
    CATCHPOINT_ANALYSER = 'catchpoint-analyser'
    CHROME = 'chrome'
    CHROMIUM = 'chromium'
    COSMOS_CRAWLER = 'cosmos-crawler'
    DOTCOM_MONITOR_BOT = 'dotcom-monitor-bot'
    EDGE = 'edge'
    FIREFOX = 'firefox'
    FIREFOX_FOCUS = 'firefox-focus'
    GOOGLE_APP_ENGINE = 'google-app-engine-software'
    INTERNET_ARCHIVER_BOT = 'internet-archiver-bot'
    INTERNET_CHANNEL = 'internet-channel'
    INTERNET_TV = 'internet-tv-browser'
    K_MELEON = 'k-meleon'
    NETSCAPE_NAVIGATOR = 'netscape-navigator'
    NETSURF = 'netsurf'
    NOKIA = 'nokia-browser'
    OBIGO = 'obigo'
    OMNIWEB = 'omniweb'
    ONEBROWSER = 'onebrowser'
    OPERA = 'opera'
    OPERA_MINI = 'opera-mini'
    PALE_MOON = 'pale-moon'
    QUPZILLA = 'qupzilla'
    ROCKMELT = 'rockmelt'
    SKYFIRE = 'skyfire'
    SPEEDCURVE_SPEED_TESTER = 'speedcurve-speed-tester'
    UC_BROWSER = 'uc-browser'
    WEBKIT = 'webkit-based-browser'
    WEBTV = 'webtv'
    YANDEX = 'yandex-browser'
    YANDEX_SEARCH_BOT = 'yandex-search-bot'
    YODAOBOT_SEARCH_BOT = 'yodaobot-search-bot'
    BLAZER = 'blazer'
    AMAZON_API_GATEWAY = 'amazon-api-gateway'
    ALEXA_BOT = 'alexa-bot'
    GOOGLE_EARTH = 'google-earth'
    EVALIANT_IMPRESSIONS_BOT = 'evaliant-impressions-bot'
    PLAYBOOK_WEB_BROWSER = 'playbook-web-browser'
    FACEBOOK_APP = 'facebook-app'
    ELECTRON_APPLICATION = 'electron-application'
    ONE_TWENTY_SIX_BROWSER = '126-browser'
    DUCKDUCKGO_FAVICONS_BOT = 'duckduckgo-favicons-bot'
    PYTHON_URLLIB = 'python-urllib'
    NETFRONT_BROWSER_NX = 'netfront-browser-nx'
    AVANT_BROWSER = 'avant-browser'
    MERCURY_BROWSER = 'mercury-browser'
    FLOCK = 'flock'
    VALVE_STEAM_TENFOOT_DISPLAY = 'valve-steam-tenfoot-display'
    CHARLOTTE = 'charlotte'
    KINDLE_BROWSER = 'kindle-browser'
    TEASHARK = 'teashark'
    GOMEZAGENT = 'gomezagent'
    QQ_BROWSER = 'qq-browser'
    OFFICE = 'office'
    BROWSERSHOTS_COM_CROSS_BROWSER_TESTER = 'browsershots-com-cross-browser-tester'
    THUNDERBIRD = 'thunderbird'
    WYZO = 'wyzo'
    GOOGLE_FAVICON_CRAWLER = 'google-favicon-crawler'
    SECONDLIFE = 'secondlife'
    INTERNET_EXPLORER_MOBILE = 'internet-explorer-mobile'
    SONGBIRD = 'songbird'
    EIGHTY_LEGS_WEB_CRAWLER = '80legs-web-crawler'
    SQUIDER_BOT = 'squider-bot'
    SONY_WEB_BROWSER = 'sony-web-browser'
    MAXTHON = 'maxthon'
    FAST_ENTERPRISE_CRAWLER = 'fast-enterprise-crawler'
    WORD = 'word'
    ACCOONA_BUSINESS_CRAWLER = 'accoona-business-crawler'
    WEBVIEW_BASED_BROWSER = 'webview-based-browser'
    PLAYSTATION_4_BROWSER = 'playstation-4-browser'
    INTERNET_EXPLORER = 'internet-explorer'
    ELINKS = 'elinks'
    SLIMBROWSER = 'slimbrowser'
    ONENOTE = 'onenote'
    DIFFBOT_SCANNER = 'diffbot-scanner'
    DELPHI_EMBEDDED_WEB_BROWSER = 'delphi-embedded-web-browser'
    SPRAYCAN = 'spraycan'
    GOOGLE_EARTH_PRO = 'google-earth-pro'
    TENFOURFOX = 'tenfourfox'
    DOLFIN = 'dolfin'
    NOOK_WEB_BROWSER = 'nook-web-browser'
    ALEXA_SITE_AUDIT = 'alexa-site-audit'
    MOTOROLA_INTERNET_BROWSER = 'motorola-internet-browser'
    W3C_LINK_CHECKER = 'w3c-link-checker'
    PINTEREST_BOT = 'pinterest-bot'
    SOGOU_SEARCH_DOG = 'sogou-search-dog'
    FRONTPAGE = 'frontpage'
    BROADSIGN_XPRESS = 'broadsign-xpress'
    YAHOO_SLURP_WEB_CRAWLER_BOT = 'yahoo-slurp-web-crawler-bot'
    COMODO_DRAGON = 'comodo-dragon'
    EPIPHANY = 'epiphany'
    DOTNETDOTCOMDOTORG_CRAWLER = 'dotnetdotcomdotorg-crawler'
    ICAB = 'icab'
    VIENNA = 'vienna'
    RUXITSYNTHETIC = 'ruxitsynthetic'
    OKHTTP = 'okhttp'
    GNIP_UNWINDFETCHOR_CRAWLER = 'gnip-unwindfetchor-crawler'
    SILK = 'silk'
    FACEBOOK_BOT = 'facebook-bot'
    WECHAT = 'wechat'
    MICROSOFT_CRYPTOAPI = 'microsoft-cryptoapi'
    KAZEHAKASE = 'kazehakase'
    BLUE_CHROME = 'blue-chrome'
    NETCRAFT_WEB_SERVER_SURVEY = 'netcraft-web-server-survey'
    SHIRETOKO = 'shiretoko'
    GRAPESHOT_BOT = 'grapeshot-bot'
    EXCEL = 'excel'
    QTCARBROWSER = 'qtcarbrowser'
    THREE_SIXTY_SPIDER = '360spider'
    CATCHBOT = 'catchbot'
    WEBPOSITIVE = 'webpositive'
    LIBRARY_FOR_WWW_IN_PERL = 'library-for-www-in-perl'
    COMETBIRD = 'cometbird'
    SAFARI = 'safari'
    SOGOU_EXPLORER = 'sogou-explorer'
    CODA = 'coda'
    IBROWSER = 'ibrowser'
    JAVA_RUNTIME_ENVIRONMENT = 'java-runtime-environment'
    GOOGLE_IMAGE_PROXY = 'google-image-proxy'
    NEXPLAYER = 'nexplayer'
    ALIENBLUE = 'alienblue'
    MIDORI = 'midori'
    W3C_VALIDATOR = 'w3c-validator'
    PUFFIN = 'puffin'
    MAJESTIC_12_DISTRIBUTED_SEARCH_BOT = 'majestic-12-distributed-search-bot'
    ICEWEASEL = 'iceweasel'
    JAVAFX_PLATFORM = 'javafx-platform'
    CONTENT_CRAWLER_SPIDER = 'content-crawler-spider'
    COVARIO_SPIDER = 'covario-spider'
    EVE_IN_GAME_BROWSER = 'eve-in-game-browser'
    OVI = 'ovi'
    BLITZBOT_CRAWLER = 'blitzbot-crawler'
    GOOGLE_SEARCH_APP = 'google-search-app'
    BAIDU_SPIDER = 'baidu-spider'
    RAPTR = 'raptr'
    FIREBIRD = 'firebird'
    ACCOONA_AI_CRAWLER = 'accoona-ai-crawler'
    WEBOS_BROWSER = 'webos-browser'
    LINKS = 'links'
    SAMSUNG_BROWSER = 'samsung-browser'
    GALEON = 'galeon'
    CLAMAV_WEBSITE_SCANNER = 'clamav-website-scanner'
    EXB_LANGUAGE_CRAWLER = 'exb-language-crawler'
    OUTLOOK = 'outlook'
    BONECHO = 'bonecho'
    NINTENDO_BROWSER = 'nintendo-browser'
    NEWSBOT = 'newsbot'
    VIVALDI = 'vivaldi'
    VISION_MOBILE_BROWSER = 'vision-mobile-browser'
    EARTHCOM_CRAWLER = 'earthcom-crawler'
    ADOBE_AIR = 'adobe-air'
    BINGPREVIEW = 'bingpreview'
    WEBLINK_PREVIEW = 'weblink-preview'
    LYNX = 'lynx'
    COMODO_ICEDRAGON = 'comodo-icedragon'
    ICEAPE = 'iceape'
    FINDLINKS_CRAWLER_BOT = 'findlinks-crawler-bot'
    MINEFIELD = 'minefield'
    NAENARA = 'naenara'
    MSN_BOT = 'msn-bot'
    CODER_NUT = 'coder-nut'
    W3M = 'w3m'
    BIT_LY_LINK_CHECKER = 'bit-ly-link-checker'
    OPENSITEEXPLORER_CRAWLER = 'opensiteexplorer-crawler'
    BAIDU_IMAGE_SPIDER = 'baidu-image-spider'
    TWITTERBOT = 'twitterbot'
    EARTHWORM = 'earthworm'
    QT_BASED_BROWSER = 'qt-based-browser'
    THEWORLD_BROWSER = 'theworld-browser'
    FLASHFIRE = 'flashfire'
    BRAVE = 'brave'
    DISCOVERY_ENGINE_CRAWLER = 'discovery-engine-crawler'
    PECL_HTTP_LIBRARY = 'pecl-http-library'
    BESLIST_SHOPPING_CRAWLER = 'beslist-shopping-crawler'
    AVANTGO_BROWSER = 'avantgo-browser'
    GOOGLE_SNIPPET_FETCHER = 'google-snippet-fetcher'
    QUALYS_SSL_ASSESSMENT_SCANNER = 'qualys-ssl-assessment-scanner'
    OPENWAVE_MOBILE_BROWSER = 'openwave-mobile-browser'
    DOOBLE = 'dooble'
    COOLNOVO = 'coolnovo'
    ADMANTX_PLATFORM_SEMANTIC_ANALYZER = 'admantx-platform-semantic-analyzer'
    CAMINO = 'camino'
    FENNEC = 'fennec'
    SAFEPAY = 'safepay'
    BROWSER_1337 = '1337browser'
    SPLASH = 'splash'
    ARORA = 'arora'
    DRAGON = 'dragon'
    NETCAST = 'netcast'
    TENCENTTRAVELER = 'tencenttraveler'
    ITUNES = 'itunes'
    EMBEDDED_WEB_BROWSER = 'embedded-web-browser'
    BOITHO_DISTRIBUTED_CRAWLER = 'boitho-distributed-crawler'
    DOMAINTOOLS_SURVEYBOT = 'domaintools-surveybot'
    BAIDU_UNION_SPIDER = 'baidu-union-spider'
    WEBLINK = 'weblink'
    PINTEREST_APP = 'pinterest-app'
    CHROMEPLUS = 'chromeplus'
    VOILABOT_BETA = 'voilabot-beta'
    ORCA = 'orca'
    LUNASCAPE = 'lunascape'
    A_PASSION_FOR_JAZZ_MEDIA_CRAWLER = 'a-passion-for-jazz-media-crawler'
    NETFRONT = 'netfront'
    GOOGLE_SITE_VERIFIER_BOT = 'google-site-verifier-bot'
    ONE_PASSWORD = '1password'
    TOPSY_BUTTERFLY_ROBOT = 'topsy-butterfly-robot'
    CAREERBOT_SEARCH_CRAWLER = 'careerbot-search-crawler'
    OUTFORM_DIGITAL_DISPLAY = 'outform-digital-display'
    MSN_MEDIA_BOT = 'msn-media-bot'
    SRWARE_IRON = 'srware-iron'
    FAST_FRESH_CRAWLER = 'fast-fresh-crawler'
    JAKARTA_COMMONS_HTTPCLIENT = 'jakarta-commons-httpclient'
    VALVE_STEAM_GAME_OVERLAY = 'valve-steam-game-overlay'
    DATAPARKSEARCH_ENGINE = 'dataparksearch-engine'
    PROXIMIC_SEARCH = 'proximic-search'
    WATERFOX = 'waterfox'
    GOOGLE_STRUCTURED_DATA_TESTING_TOOL = 'google-structured-data-testing-tool'
    TWEETMEME_BOT = 'tweetmeme-bot'
    CURL = 'curl'
    VOILABOT = 'voilabot'
    AMAYA = 'amaya'
    ROCCAT = 'roccat'
    WGET = 'wget'
    GOOGLEBOT = 'googlebot'
    EA_ORIGIN_BROWSER = 'ea-origin-browser'
    SLACKBOT_LINK_CHECKER = 'slackbot-link-checker'
    REKONQ = 'rekonq'
    SEAMONKEY = 'seamonkey'
    TWITTER_APP = 'twitter-app'
    EXALEAD_CRAWLER = 'exalead-crawler'
    GOOGLE_WEBLIGHT_PROXY = 'google-weblight-proxy'
    KONQUEROR = 'konqueror'
    DOMAIN_RE_ANIMATOR_BOT = 'domain-re-animator-bot'
    LINKCHECK_ANALYSER = 'linkcheck-analyser'
    BAIDU_BOX_APP = 'baidu-box-app'
    LIGHT = 'light'
    PRO_ENGINEER_WILDFIRE = 'pro-engineer-wildfire'
    OPEN_WEBKIT_SHARP_BASED_BROWSER = 'open-webkit-sharp-based-browser'
    QIHOO_360 = 'qihoo-360'
    GENIEO_BOT = 'genieo-bot'
    SLEIPNIR = 'sleipnir'
    CHIMERA = 'chimera'
    MOSAIC = 'mosaic'
    POWERPOINT = 'powerpoint'
    PANSCIENT_CRAWLER = 'panscient-crawler'
    W3C_CSS_VALIDATOR = 'w3c-css-validator'
    UNKNOWN_BROWSER = 'unknown-browser'
    AOL_BROWSER = 'aol-browser'
    SOSOSPIDER_SEARCH_BOT = 'sosospider-search-bot'
    GOOGLEBOT_MOBILE = 'googlebot-mobile'
    IRIS = 'iris'
    BINGBOT = 'bingbot'
    GTMETRIX_ANALYSER = 'gtmetrix-analyser'
    DISCORD_BOT = 'discord-bot'
    DILLO = 'dillo'
    GOOGLE_S_MEDIA_PARTNERS_SYSTEM_ADSENSE = 'google-s-media-partners-system-adsense'
    HTTPCLIENT = 'httpclient'
    DLINK_BACKDOOR = 'dlink-backdoor'
    TELEGRAMBOT = 'telegrambot'
    QQDOWNLOAD_DOWNLOAD_MANAGER = 'qqdownload-download-manager'
    LB_BROWSER = 'lb-browser'
    AHREFS_BACKLINK_RESEARCH_BOT = 'ahrefs-backlink-research-bot'
    PHOENIX = 'phoenix'
    BECOME_COM_CRAWLER = 'become-com-crawler'


class OperatingSystem(Enum):
    UNIX = 'a-unix-based-os'
    ANDROID = 'android'
    BADA = 'bada'
    BEOS = 'beos'
    BLACKBERRY = 'blackberry-os'
    CHROMEOS = 'chromeos'
    FIRE_OS = 'fire-os'
    FREEBSD = 'freebsd'
    HAIKU = 'haiku'
    IOS = 'ios'
    LINUX = 'linux'
    MAC = 'mac'
    MAC_OS_X = 'mac-os-x'
    MACOS = 'macos'
    OPENBSD = 'openbsd'
    SUNOS = 'sunos'
    SYMBIAN = 'symbian'
    WEBOS = 'webos'
    WINDOWS = 'windows'
    WINDOWS_MOBILE = 'windows-mobile'
    WINDOWS_PHONE = 'windows-phone'
    HP_WEBOS = 'hp-webos'
    DARWIN = 'darwin'
    IRIX = 'irix'
    RIM_TABLET_OS = 'rim-tablet-os'
    LIVEAREA = 'livearea'
    PALMOS = 'palmos'


class SoftwareEngine(Enum):
    KHTML = 'khtml'
    GOANNA = 'goanna'
    EDGEHTML = 'edgehtml'
    TRIDENT = 'trident'
    PRESTO = 'presto'
    BLINK = 'blink'
    GECKO = 'gecko'
    WEBKIT = 'webkit'
    NETFRONT = 'netfront'


class HardwareType(Enum):
    COMPUTER = 'computer'
    LARGE_SCREEN__GAME_CONSOLE = 'large-screen -> game-console'
    LARGE_SCREEN__MEDIA_PLAYER = 'large-screen -> media-player'
    LARGE_SCREEN__TV = 'large-screen -> tv'
    MOBILE = 'mobile'
    MOBILE__EBOOK_READER = 'mobile -> ebook-reader'
    MOBILE__HANDHELD_GAME = 'mobile -> handheld-game'
    MOBILE__MUSIC_PLAYER = 'mobile -> music-player'
    MOBILE__PHONE = 'mobile -> phone'
    MOBILE__TABLET = 'mobile -> tablet'
    SERVER = 'server'
    PDA = 'pda'
    VEHICLE__CAR = 'vehicle -> car'
    LARGE_SCREEN = 'large-screen'
    MOBILE__WEARABLE__GLASSES = 'mobile -> wearable -> glasses'
    LARGE_SCREEN__BILLBOARD = 'large-screen -> billboard'


class SoftwareType(Enum):
    APPLICATION__SOFTWARE_LIBRARY = 'application -> software-library'
    BOT__ANALYSER = 'bot -> analyser'
    BOT__CRAWLER = 'bot -> crawler'
    BOT__SITE_MONITOR = 'bot -> site-monitor'
    WEB_BROWSER = 'browser -> web-browser'
    APPLICATION__MEDIA_PLAYER = 'application -> media-player'
    APPLICATION__DOWNLOAD_HELPER = 'application -> download-helper'
    APPLICATION = 'application'
    BOT__SECURITY_ANALYSER = 'bot -> security-analyser'
    BROWSER__IN_APP_BROWSER = 'browser -> in-app-browser'
    BOT__TOOL = 'bot -> tool'
    APPLICATION__PROXY = 'application -> proxy'
    APPLICATION__BILLBOARD = 'application -> billboard'


class Popularity(Enum):
    POPULAR = "very common"
    COMMON = "common"
    AVERAGE = "average"
    UNCOMMON = "uncommon"



# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from subprocess import DEVNULL

import errno
import os
import subprocess
from platform import system
from subprocess import PIPE
from time import sleep
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.common import utils


_HAS_NATIVE_DEVNULL = True


class Service(object):

    def __init__(self, executable, port=0, log_file=DEVNULL, env=None, start_error_message=""):
        self.path = executable

        self.port = port
        if self.port == 0:
            self.port = utils.free_port()

        if not _HAS_NATIVE_DEVNULL and log_file == DEVNULL:
            log_file = open(os.devnull, 'wb')

        self.start_error_message = start_error_message
        self.log_file = log_file
        # Default value for every python subprocess: subprocess.Popen(..., creationflags=0)
        self.creationflags = 0
        self.env = env or os.environ

    @property
    def service_url(self):
        """
        Gets the url of the Service
        """
        return "http://%s" % utils.join_host_port('localhost', self.port)

    def command_line_args(self):
        raise NotImplementedError("This method needs to be implemented in a sub class")

    def start(self):
        """
        Starts the Service.

        :Exceptions:
         - WebDriverException : Raised either when it can't start the service
           or when it can't connect to the service
        """
        try:
            cmd = [self.path]
            cmd.extend(self.command_line_args())
            self.process = subprocess.Popen(cmd, env=self.env,
                                            close_fds=system() != 'Windows',
                                            stdout=self.log_file,
                                            stderr=self.log_file,
                                            stdin=PIPE,
                                            creationflags=self.creationflags)
        except TypeError:
            raise
        except OSError as err:
            if err.errno == errno.ENOENT:
                raise WebDriverException(
                    "'%s' executable needs to be in PATH. %s" % (
                        os.path.basename(self.path), self.start_error_message)
                )
            elif err.errno == errno.EACCES:
                raise WebDriverException(
                    "'%s' executable may have wrong permissions. %s" % (
                        os.path.basename(self.path), self.start_error_message)
                )
            else:
                raise
        except Exception as e:
            raise WebDriverException(
                "The executable %s needs to be available in the path. %s\n%s" %
                (os.path.basename(self.path), self.start_error_message, str(e)))
        count = 0
        while True:
            self.assert_process_still_running()
            if self.is_connectable():
                break

            count += 1
            sleep(0.5)
            if count == 60:
                raise WebDriverException("Can not connect to the Service %s" % self.path)

    def assert_process_still_running(self):
        return_code = self.process.poll()
        if return_code:
            raise WebDriverException(
                'Service %s unexpectedly exited. Status code was: %s'
                % (self.path, return_code)
            )

    def is_connectable(self):
        return utils.is_connectable(self.port)

    def send_remote_shutdown_command(self):
        from urllib import request as url_request
        URLError = url_request.URLError

        try:
            url_request.urlopen("%s/shutdown" % self.service_url)
        except URLError:
            return

        for x in range(30):
            if not self.is_connectable():
                break
            else:
                sleep(1)

    def stop(self):
        """
        Stops the service.
        """
        if self.log_file != PIPE and not (self.log_file == DEVNULL and _HAS_NATIVE_DEVNULL):
            try:
                self.log_file.close()
            except Exception:
                pass

        if not self.process:
            return

        try:
            self.send_remote_shutdown_command()
        except TypeError:
            pass

        try:
            if self.process:
                for stream in [self.process.stdin,
                               self.process.stdout,
                               self.process.stderr]:
                    try:
                        stream.close()
                    except AttributeError:
                        pass
                self.process.terminate()
                self.process.wait()
                self.process.kill()
                self.process = None
        except OSError:
            pass

    def __del__(self):
        # `subprocess.Popen` doesn't send signal on `__del__`;
        # so we attempt to close the launched process when `__del__`
        # is triggered.
        try:
            self.stop()
        except Exception:
            pass



# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from typing import Dict, List, Union
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement


def with_tag_name(tag_name: str) -> "RelativeBy":
    """
        Start searching for relative objects using a tag name.

        Note: This method may be removed in future versions, please use
        `locate_with` instead.
        :Args:
            - tag_name: the DOM tag of element to start searching.
        :Returns:
            - RelativeBy - use this object to create filters within a
                `find_elements` call.
    """
    if not tag_name:
        raise WebDriverException("tag_name can not be null")
    return RelativeBy({"css selector": tag_name})


def locate_with(by: By, using: str) -> "RelativeBy":
    """
        Start searching for relative objects your search criteria with By.

        :Args:
            - by: The value from `By` passed in.
            - using: search term to find the element with.
        :Returns:
            - RelativeBy - use this object to create filters within a
                `find_elements` call.
    """
    assert by is not None, "Please pass in a by argument"
    assert using is not None, "Please pass in a using argument"
    return RelativeBy({by: using})


class RelativeBy(object):
    """
        Gives the opportunity to find elements based on their relative location
        on the page from a root elelemt. It is recommended that you use the helper
        function to create it.

        Example:
            lowest = driver.find_element(By.ID, "below")

            elements = driver.find_elements(locate_with(By.CSS_SELECTOR, "p").above(lowest))

            ids = [el.get_attribute('id') for el in elements]
            assert "above" in ids
            assert "mid" in ids
    """

    def __init__(self, root: Dict[By, str] = None, filters: List = None):
        """
            Creates a new RelativeBy object. It is preferred if you use the
            `locate_with` method as this signature could change.
            :Args:
                root - A dict with `By` enum as the key and the search query as the value
                filters - A list of the filters that will be searched. If none are passed
                    in please use the fluent API on the object to create the filters
        """
        self.root = root
        self.filters = filters or []

    def above(self, element_or_locator: Union[WebElement, Dict] = None) -> "RelativeBy":
        """
            Add a filter to look for elements above.
            :Args:
                - element_or_locator: Element to look above
        """
        if not element_or_locator:
            raise WebDriverException("Element or locator must be given when calling above method")

        self.filters.append({"kind": "above", "args": [element_or_locator]})
        return self

    def below(self, element_or_locator: Union[WebElement, Dict] = None) -> "RelativeBy":
        """
            Add a filter to look for elements below.
            :Args:
                - element_or_locator: Element to look below
        """
        if not element_or_locator:
            raise WebDriverException("Element or locator must be given when calling above method")

        self.filters.append({"kind": "below", "args": [element_or_locator]})
        return self

    def to_left_of(self, element_or_locator: Union[WebElement, Dict] = None) -> "RelativeBy":
        """
            Add a filter to look for elements to the left of.
            :Args:
                - element_or_locator: Element to look to the left of
        """
        if not element_or_locator:
            raise WebDriverException("Element or locator must be given when calling above method")

        self.filters.append({"kind": "left", "args": [element_or_locator]})
        return self

    def to_right_of(self, element_or_locator: Union[WebElement, Dict] = None) -> "RelativeBy":
        """
            Add a filter to look for elements right of.
            :Args:
                - element_or_locator: Element to look right of
        """
        if not element_or_locator:
            raise WebDriverException("Element or locator must be given when calling above method")

        self.filters.append({"kind": "right", "args": [element_or_locator]})
        return self

    def near(self, element_or_locator_distance: Union[WebElement, Dict, int] = None) -> "RelativeBy":
        """
            Add a filter to look for elements above.
            :Args:
                - element_or_locator_distance: Element to look near by the element or within a distance
        """
        if not element_or_locator_distance:
            raise WebDriverException("Element or locator or distance must be given when calling above method")

        self.filters.append({"kind": "near", "args": [element_or_locator_distance]})
        return self

    def to_dict(self) -> Dict:
        """
            Create a dict that will be passed to the driver to start searching for the element
        """
        return {
            'relative': {
                'root': self.root,
                'filters': self.filters,
            }
        }



# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.common.options import ArgOptions


class Log(object):
    def __init__(self):
        self.level = None

    def to_capabilities(self) -> dict:
        if self.level:
            return {"log": {"level": self.level}}
        return {}


class Options(ArgOptions):
    KEY = "safari.options"

    def __init__(self):
        super(Options, self).__init__()
        self._binary_location = None
        self._preferences: dict = {}
        self.log = Log()

    @property
    def binary_location(self) -> str:
        """
        :Returns: The location of the browser binary otherwise an empty string
        """
        return self._binary_location

    @binary_location.setter
    def binary_location(self, value: str):
        """
        Allows you to set the browser binary to launch

        :Args:
         - value : path to the browser binary
        """
        self._binary_location = value

    def to_capabilities(self) -> dict:
        """Marshals the  options to an desired capabilities object.
        """
        # This intentionally looks at the internal properties
        # so if a binary or profile has _not_ been set,
        # it will defer to geckodriver to find the system Firefox
        # and generate a fresh profile.
        caps = self._caps
        opts = {}

        if self._arguments:
            opts["args"] = self._arguments
        if self._binary_location:
            opts["binary"] = self._binary_location
        opts.update(self.log.to_capabilities())

        if opts:
            caps[Options.KEY] = opts

        return caps

    @property
    def default_capabilities(self) -> dict:
        return DesiredCapabilities.SAFARI.copy()



import time
import datetime
import pytz


def get_cur_date():
    return datetime.datetime.now(pytz.timezone('PRC')).strftime("%Y-%m-%d")


def get_cur_datetime():
    return datetime.datetime.now(pytz.timezone('PRC')).strftime("%Y-%m-%d %H:%M:%S")


def ns_timestamp_to_datetime(timestamp):
    try:
        return str(datetime.datetime.fromtimestamp(int(timestamp / 1000)))
    except Exception:
        return ""


def timestamp_to_datetime(timestamp):
    try:
        return str(datetime.datetime.fromtimestamp(int(timestamp)))
    except Exception:
        return ""


def datetime_to_timestamp(dt):
    if not dt:
        return
    try:
        timeArray = time.strptime(dt, "%Y-%m-%d %H:%M:%S")
        # 转换成时间戳
        timestamp = int(time.mktime(timeArray))
        return timestamp
    except Exception:
        pass

def datetime_to_format(time_string,entry_format,output_format):
    try:
        timeArray = time.strptime(time_string,entry_format)
        t  = time.strftime(output_format,timeArray)
        return t
    except Exception:
        return time_string



def datetime_to_datetimezone(dt,zone="00:00"):
    if not dt:
        return
    try:
        timeArray = time.strptime(dt, "%Y-%m-%d %H:%M:%S")
        # 转换时区格式
        zone_time = time.strftime("%Y-%m-%dT%H:%M:%S+{}".format(zone),timeArray)
        return zone_time
    except Exception:
        pass

if __name__ == '__main__':
    pass


# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import List
from selenium.webdriver.chromium import service


DEFAULT_EXECUTEABLE_PATH = "chromedriver"


class Service(service.ChromiumService):
    """
    Object that manages the starting and stopping of the ChromeDriver
    """

    def __init__(self, executable_path: str = DEFAULT_EXECUTEABLE_PATH,
                 port: int = 0, service_args: List[str] = None,
                 log_path: str = None, env: dict = None):
        """
        Creates a new instance of the Service

        :Args:
         - executable_path : Path to the ChromeDriver
         - port : Port the service is running on
         - service_args : List of args to pass to the chromedriver service
         - log_path : Path for the chromedriver service to log to"""

        super(Service, self).__init__(
            executable_path,
            port,
            service_args,
            log_path,
            env,
            "Please see https://chromedriver.chromium.org/home")



import json
import os
import random
import zipfile


class UserAgent:
    ATTRIBUTES_MAP = {
        'hardware_types': [],
        'software_types': [],
        'software_names': [],
        'software_engines': [],
        'operating_systems': [],
        'popularity': [],
    }

    def __init__(self, limit=None, *args, **kwargs):
        self.user_agents = []

        for attribute, values in self.ATTRIBUTES_MAP.items():
            setattr(self, attribute, kwargs.get(attribute, [v.lower() for v in values]))

        for user_agent in self.load_user_agents():

            if limit is not None and len(self.user_agents) >= limit:
                break

            if self.hardware_types and user_agent['hardware_type'].lower() not in self.hardware_types:
                continue

            if self.software_types and user_agent['software_type'].lower() not in self.software_types:
                continue

            if self.software_names and user_agent['software_name'].lower() not in self.software_names:
                continue

            if self.software_engines and user_agent['software_engine'].lower() not in self.software_engines:
                continue

            if self.operating_systems and user_agent['operating_system'].lower() not in self.operating_systems:
                continue

            if self.popularity and user_agent['popularity'].lower() not in self.popularity:
                continue

            self.user_agents.append(user_agent)

    def load_user_agents(self):
        file_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data/user_agents.zip')

        with zipfile.ZipFile(file_path) as zipped_user_agents:
            with zipped_user_agents.open('user_agents.jl') as user_agents:
                for user_agent in user_agents:

                    if hasattr(user_agent, 'decode'):
                        user_agent = user_agent.decode()

                    yield json.loads(user_agent)

    def get_user_agents(self):
        return self.user_agents

    def get_random_user_agent(self):
        return random.choice(self.user_agents)['user_agent']



from similib.cmd import exe_command
exe_command("rm -rf ./dist/*")
exe_command("pip install --upgrade setuptools wheel")
exe_command("python setup.py sdist bdist_wheel")
exe_command("pip install --upgrade twine")
# exe_command("python -m twine upload --repository-url https://upload.pypi.org/legacy/ dist/*")


"""
mongodb数据文件操作
"""
from similib.cmd import exe_command
import pymongo


class Mongo(object):
    def __init__(self, ip, port, user, passwd):
        self.ip = ip
        self.port = port
        self.user = user
        self.passwd = passwd
        self.conn = pymongo.MongoClient(host=ip, port=port, username=user, password=passwd)

    def import_json(self, db, col, file_path):
        cmd = "mongoimport -h {} --port {} -u {} -p {} --authenticationDatabase admin --db {} --collection {} --file {}".format(
            self.ip, self.port, self.user, self.passwd, db, col, file_path)
        exe_command(cmd)

    def export_json(self, db, col, file_path):
        cmd = "mongoexport -h {} --port {} -u {} -p {} --authenticationDatabase admin -d {} -c {} -o {}".format(
            self.ip, self.port, self.user, self.passwd, db, col, file_path)
        exe_command(cmd)

    def import_csv(self, db, col, file_path):
        cmd = "mongoimport -h {} --port {} -u {} -p {} --authenticationDatabase admin --db {} --collection {} --type csv --headerline --ignoreBlanks --file {}".format(
            self.ip, self.port, self.user, self.passwd, db, col, file_path)
        exe_command(cmd)

    def export_csv(self, db, col, file_path):
        cmd = "mongoexport -h {} --port {} -u {} -p {} --authenticationDatabase admin -d {} -c {} --type csv -fields --out {} --limit 50".format(
            self.ip, self.port, self.user, self.passwd, db, col, file_path)
        exe_command(cmd)

    def create_index(self, db, col, fields: list, direct=pymongo.ASCENDING, raw=None, unique=True):
        raw = raw if raw else []
        for f in fields:
            raw.append((f, direct))
        self.conn[db][col].create_index(raw, unique=unique)


if __name__ == '__main__':
    pass



from selenium.common.exceptions import NoSuchElementException

from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
import time


def wait_ele(driver, xpath, ts=10):
    try:
        WebDriverWait(driver, ts).until(
            EC.presence_of_element_located((By.XPATH, xpath))
        )
    except Exception:
        pass


def my_wait(driver, xpath, ts=10):
    for i in range(ts):
        try:
            driver.find_element_by_xpath(xpath)
        except Exception:
            time.sleep(1)
        else:
            return True
    return False


def driver_safe_find_element_by_xpath(driver, ele_xpath):
    try:
        return driver.find_element_by_xpath(ele_xpath)
    except NoSuchElementException:
        return None


def driver_safe_find_elements_by_xpath(driver, ele_xpath):
    try:
        return driver.find_elements_by_xpath(ele_xpath)
    except NoSuchElementException:
        return None


def drive_wait_element_by_xpath(driver, ele_xpath, ts=10):
    my_wait(driver, ele_xpath, ts)
    return driver_safe_find_element_by_xpath(driver, ele_xpath)


def ele_safe_find_element_by_xpath(ele, ele_xpath):
    try:
        return ele.find_element_by_xpath(ele_xpath)
    except NoSuchElementException:
        return


def isclickable(driver,xpath):
    try:
        WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable((By.XPATH, xpath)))
        return True
    except :
        return False


if __name__ == '__main__':
   pass



# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from selenium.webdriver.remote.remote_connection import RemoteConnection


class ChromiumRemoteConnection(RemoteConnection):
    def __init__(self, remote_server_addr, vendor_prefix, browser_name, keep_alive=True, ignore_proxy=False):
        RemoteConnection.__init__(self, remote_server_addr, keep_alive, ignore_proxy=ignore_proxy)
        self.browser_name = browser_name
        self._commands["launchApp"] = ('POST', '/session/$sessionId/chromium/launch_app')
        self._commands["setPermissions"] = ('POST', '/session/$sessionId/permissions')
        self._commands["setNetworkConditions"] = ('POST', '/session/$sessionId/chromium/network_conditions')
        self._commands["getNetworkConditions"] = ('GET', '/session/$sessionId/chromium/network_conditions')
        self._commands["deleteNetworkConditions"] = ('DELETE', '/session/$sessionId/chromium/network_conditions')
        self._commands['executeCdpCommand'] = ('POST', '/session/$sessionId/{}/cdp/execute'.format(vendor_prefix))
        self._commands['getSinks'] = ('GET', '/session/$sessionId/{}/cast/get_sinks'.format(vendor_prefix))
        self._commands['getIssueMessage'] = ('GET', '/session/$sessionId/{}/cast/get_issue_message'.format(vendor_prefix))
        self._commands['setSinkToUse'] = ('POST', '/session/$sessionId/{}/cast/set_sink_to_use'.format(vendor_prefix))
        self._commands['startTabMirroring'] = ('POST', '/session/$sessionId/{}/cast/start_tab_mirroring'.format(vendor_prefix))
        self._commands['stopCasting'] = ('POST', '/session/$sessionId/{}/cast/stop_casting'.format(vendor_prefix))
